package cmsc420_f22;

import java.util.ArrayList;

// ---------------------------------------------------------------------
// Author: Dave Mount
// For: CMSC 420
// Date: Fall 2022
//
// This is an implementation of a LeftistHeap data structure. This is a
// priority queue, which in addition to supporting insertion and extract
// min, it supports merging of two heaps.
// ---------------------------------------------------------------------

public class LeftistHeap<Key extends Comparable<Key>, Value> {

	// -----------------------------------------------------------------
	// Node in the tree
	// -----------------------------------------------------------------

	class LHNode {
		Key key; // key (priority)
		Value value; // value (application dependent)
		LHNode left; // children
		LHNode right;
		int npl; // null path length

		/**
		 * Basic constructor.
		 */
		LHNode(Key x, Value v, LHNode left, LHNode right, LHNode parent) {
			this.key = x;
			this.value = v;
			this.left = left;
			this.right = right;
			this.npl = 0;
		}
	}

	// -----------------------------------------------------------------
	// Private members
	// -----------------------------------------------------------------

	private LHNode root; // root of tree

	// -----------------------------------------------------------------
	// Local utilities
	// -----------------------------------------------------------------

	/**
	 * Compute a node's npl from its children
	 */
	int npl(LHNode u) {
		return (u == null ? -1 : 1 + Math.min(npl(u.left), npl(u.right)));
	}

	/**
	 * Merge helper. Merge subtrees rooted at u and v.
	 */
	LHNode merge(LHNode u, LHNode v) {
		if (u == null)
			return v; // if one is empty, return the other
		if (v == null)
			return u;
		if (u.key.compareTo(v.key) > 0) { // swap so that u is smaller
			LHNode t = u;
			u = v;
			v = t;
		}
		if (u.left == null) { // u has nothing to its left
			u.left = v; // put v here then
		} else { // merge v on right and swap if needed
			u.right = merge(u.right, v); // recursively merge u's right subtree
			if (u.left.npl < u.right.npl) { // fail the leftist property?
				LHNode t = u.left; // swap children
				u.left = u.right;
				u.right = t;
			}
			u.npl = u.right.npl + 1; // update npl value
		}
		return u; // return the root of final tree
	}

	/**
	 * Find all subtrees with keys larger than x, unlinks them, and returns them in
	 * list. The list is sorted in inorder.
	 */
	LHNode getBigs(LHNode u, Key x, ArrayList<LHNode> bigs) { 
		if (u != null) {
			if (u.key.compareTo(x) <= 0) { // small key?
				u.left = getBigs(u.left, x, bigs); // apply to children
				u.right = getBigs(u.right, x, bigs);
				return u;
			} else {
				bigs.add(u); // add this subtree to the list
				return null; // unlink from parent
			}
		} else {
			return null;
		}
	}

	/**
	 * Fix up NPL values and leftist properties throughout subtree.
	 */
	void fixUp(LHNode u) {
		if (u != null) {
			fixUp(u.left); // fix our subtrees
			fixUp(u.right);
			if (npl(u.left) < npl(u.right)) { // leftist property violated?
				LHNode temp = u.left; // swap children
				u.left = u.right;
				u.right = temp;
			}
			u.npl = 1 + npl(u.right); // our npl is our right child's npl
		}
	}

	/**
	 * List the nodes of subtree in reverse preorder. By "reverse" we mean that the
	 * right subtree comes before the left.
	 */
	ArrayList<String> reversePreorder(LHNode u) {
		ArrayList<String> list = new ArrayList<String>();
		if (u == null) {
			list.add("[]"); // null link indicator
		} else {
			list.add("(" + u.key + ", " + u.value + ") [" + u.npl + "]"); // add this node
			list.addAll(reversePreorder(u.right)); // process right
			list.addAll(reversePreorder(u.left)); // process left
		}
		return list;
	}

	// -----------------------------------------------------------------
	// Public members
	// -----------------------------------------------------------------

	/**
	 * Default constructor.
	 */
	public LeftistHeap() {
		clear();
	}

	/**
	 * Is the heap empty?
	 */
	public boolean isEmpty() {
		return root == null;
	}

	/**
	 * Clear the entire structure.
	 */
	public void clear() {
		root = null;
	}

	/**
	 * Insert key-value pair.
	 */
	public void insert(Key x, Value v) {
		LHNode u = new LHNode(x, v, null, null, null); // create single node
		root = merge(root, u); // merge it with the root
	}

	/**
	 * Merge with another heap. This clears the other heap.
	 */
	public void mergeWith(LeftistHeap<Key, Value> h2) {
		if (h2 == null || h2 == this) return; // trivial - ignore
		root = merge(root, h2.root); // merge the two trees
		h2.clear();
	}

	/**
	 * Split this heap about x. Elements less than or equal to x remain
	 * in this heap, and larger elements are added to a second heap, which
	 * is returned.
	 */
	public LeftistHeap<Key, Value> split(Key x) {
		ArrayList<LHNode> bigs = new ArrayList<LHNode>(); // list of big subtrees
		root = getBigs(root, x, bigs); // collect the big subtrees
		LeftistHeap<Key, Value> h2 = new LeftistHeap<Key, Value>(); // create second heap
		for (LHNode u : bigs) {
			h2.root = merge(h2.root, u); // merge u with h2
		}
		fixUp(root); // restore npls and leftism
		return h2;
	}

	/**
	 * Get the minimum key.
	 */
	public Key getMinKey() {
		return (root == null ? null : root.key);
	}

	/**
	 * Extract the minimum item from the heap.
	 */
	public Value extractMin() throws Exception {
		if (isEmpty()) {
			throw new Exception("Empty heap");
		}
		Value result = root.value; // final result
		root = merge(root.left, root.right); // merge subtrees
		return result;
	}

	/**
	 * Get a list of entries in reverse (right-left) preorder.
	 */
	public ArrayList<String> list() {
		ArrayList<String> list = new ArrayList<String>();
		list.addAll(reversePreorder(root));
		return list;
	}

}
